#!/usr/bin/env python3
import json
import unittest

from jinsi import render_json


class JinsiExamples(unittest.TestCase):

    def test_simple_template(self):
        doc = """
            ::let:
              user:
                ::object:
                  - ::titlecase:
                      ::get: $user.username
                  - Type: AWS::IAM::User
                    Properties:
                      UserName:
                        ::get: $user.username
                      Groups:
                        - Administrators
                      LoginProfile:
                        Password:
                          ::get: $user.password
                          ::else: default
                        PasswordResetRequired: Yes
              users:
                ::merge:
                  ::each $ as $user:
                    ::call user:
            
            Resources:
              ::call users:
                - username: jim
                  password: one
                - username: jack
                  password: two
                - username: johnny
        """
        rendered = render_json(doc)
        self.assertEqual({
            "Resources": {
                "Jim": {
                    "Type": "AWS::IAM::User",
                    "Properties": {
                        "UserName": "jim",
                        "Groups": ["Administrators"],
                        "LoginProfile": {
                            "Password": "one",
                            "PasswordResetRequired": True
                        }
                    }
                },
                "Jack": {
                    "Type": "AWS::IAM::User",
                    "Properties": {
                        "UserName": "jack",
                        "Groups": ["Administrators"],
                        "LoginProfile": {
                            "Password": "two",
                            "PasswordResetRequired": True
                        }
                    }
                },
                "Johnny": {
                    "Type": "AWS::IAM::User",
                    "Properties": {
                        "UserName": "johnny",
                        "Groups": ["Administrators"],
                        "LoginProfile": {
                            "Password": "default",
                            "PasswordResetRequired": True
                        }
                    }
                }
            }
        }, json.loads(rendered))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import os.path
from typing import Union

import yaml

from .environment import Environment
from .jsonutil import dumpjson
from .nodes import Node, Empty, Value, Constant
from .parser import Parser
from .yamlutil import Loader, Dumper, dumpyaml

# noinspection PyBroadException
try:
    PATH_MAX = os.pathconf('/', 'PC_PATH_MAX')
except Exception:
    PATH_MAX = 1024


def parse(doc: Union[str, dict, list]) -> Node:
    if isinstance(doc, str):
        return parse_yaml(doc)
    return Parser().parse_node(doc, Empty())


def parse_yaml(yaml_str: str) -> Node:
    import textwrap
    yaml_str = textwrap.dedent(yaml_str)
    doc = yaml.load(yaml_str, Loader=Loader)
    if not isinstance(doc, (list, dict)):
        return Constant(Empty(), doc)
    return parse(doc)


def parse_file(path: str) -> Node:
    with open(path) as f:
        doc = yaml.load(f, Loader=Loader)
    if not isinstance(doc, (list, dict)):
        return Constant(Empty(), doc)
    return parse(doc)


def parse_from(file_like) -> Node:
    doc = yaml.load(file_like, Loader=Loader)
    if not isinstance(doc, (list, dict)):
        return Constant(Empty(), doc)
    return parse(doc)


def evaluate(doc: Union[Node, str, dict, list], **env) -> Value:
    if not isinstance(doc, Node):
        doc: Node = parse(doc)
    return doc.evaluate(Environment(**env))


def render_yaml(doc: Union[Node, str, dict, list], **env) -> str:
    result = evaluate(doc, **env)
    return dumpyaml(result)


def render_file_yaml(file: str, **env) -> str:
    node = parse_file(file)
    result = evaluate(node, **env)
    return dumpyaml(result)


def render_json(doc: Union[Node, str, dict, list], **env) -> str:
    result = evaluate(doc, **env)
    return dumpjson(result)


def render_file_json(file: str, **env) -> str:
    node = parse_file(file)
    result = evaluate(node, **env)
    return dumpjson(result)


def load_yaml(yaml_str: str, **env) -> Value:
    doc = parse_yaml(yaml_str)
    result = evaluate(doc, **env)
    return result


def load_file(path: str, **env) -> Value:
    doc = parse_file(path)
    result = evaluate(doc, **env)
    return result


def load_from(file_like, **env) -> Value:
    doc = parse_from(file_like)
    result = evaluate(doc, **env)
    return result


def load(thing, **env) -> Value:
    if isinstance(thing, str):
        if len(thing) <= PATH_MAX and os.path.exists(thing):
            return load_file(thing, **env)
        return load_yaml(thing, **env)
    return load_from(thing, **env)



from __future__ import annotations

import functools
import hashlib
import re
import struct
from decimal import Decimal
from typing import Callable, List, Optional

from dezimal import Dec

from jinsi.exceptions import NoMergePossible


def hash_complex(
        value, *, algo="sha256",
        _prefix: Optional[bytes] = None,
        _none=bytes([3]),
        _true=bytes([5]),
        _false=bytes([7]),
        _zero=bytes([11]),
        _int_neg=bytes([13]),
        _int_pos=bytes([17]),
        _double=bytes([19]),
        _string=bytes([23]),
        _seq=bytes([29]),
        _map=bytes([31]),
        _set=bytes([37]),
        _arbitrary=bytes([41]),
        _custom=bytes([43]),
) -> bytes:
    if isinstance(value, str):
        md = hashlib.new(algo)
        md.update(_string)
        md.update(value.encode('utf8'))
        return md.digest()
    if isinstance(value, bool):
        md = hashlib.new(algo)
        if value:
            md.update(_true)
        else:
            md.update(_false)
        return md.digest()
    if isinstance(value, int):
        md = hashlib.new(algo)
        if value == 0:
            md.update(_zero)
        elif value < 0:
            value *= -1
            md.update(_int_neg)
        else:
            md.update(_int_pos)
        while value > 0:
            md.update(bytes([value % 256]))
            value //= 256
        return md.digest()
    if isinstance(value, float):
        md = hashlib.new(algo)
        md.update(_double)
        md.update(struct.pack("d", value))
        return md.digest()
    if isinstance(value, (list, tuple)):
        md = hashlib.new(algo)
        md.update(_seq)
        for item in value:
            md.update(hash_complex(item, algo=algo))
        return md.digest()
    if isinstance(value, (set, frozenset)):
        md = hashlib.new(algo)
        md.update(_set)
        hashes = []
        for item in value:
            hashes.append(hash_complex(item, algo=algo))
        hashes.sort()
        for h in hashes:
            md.update(h)
        return md.digest()
    if isinstance(value, dict):
        md = hashlib.new(algo)
        if _prefix is None:
            md.update(_map)
        else:
            md.update(_prefix)
        _seq = []
        for key, value in value.items():
            _seq.append((hash_complex(key, algo=algo), hash_complex(value, algo=algo)))
        _seq.sort()
        for key, value in _seq:
            md.update(key)
            md.update(value)
        return md.digest()
    if isinstance(value, (bytes, bytearray)):
        md = hashlib.new(algo)
        md.update(_arbitrary)
        md.update(value)
        return md.digest()
    if value is None:
        md = hashlib.new(algo)
        md.update(_none)
        return md.digest()
    if hasattr(value, '__class__'):
        prefix = bytearray()
        prefix.extend(_custom)
        prefix.extend(hash_complex(value.__class__.__module__, algo=algo))
        prefix.extend(hash_complex(value.__class__.__name__, algo=algo))
        return hash_complex(value.__dict__, algo=algo, _prefix=prefix)
    raise ValueError(value)


def cached_function(func):
    _empty_args_hash = hash_complex(((), {}), algo='sha1').hex()
    _cache = {}

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not args and not kwargs:
            args_hash = _empty_args_hash
        else:
            args_hash = hash_complex((args, kwargs), algo='sha1').hex()
        try:
            result = _cache[args_hash]
        except KeyError:
            result = func(*args, **kwargs)
            _cache[args_hash] = result
        return result

    return wrapper


def cached_method(func):
    _empty_args_hash = hash_complex(((), {}), algo='sha1').hex()

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            cache = self.__method_cache__
        except AttributeError:
            self.__method_cache__ = {}
            cache = self.__method_cache__
        if not args and not kwargs:
            args_hash = _empty_args_hash
        else:
            args_hash = hash_complex((args, kwargs), algo='sha1').hex()
        name = f"{func.__name__}:{args_hash}"
        try:
            result = cache[name]
        except KeyError:
            result = func(self, *args, **kwargs)
            cache[name] = result
        return result

    return wrapper


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def select(obj, *path, fallback=None):
    current = obj
    for element in path:
        try:
            current = current[element]
        except (KeyError, TypeError):
            return fallback
    return current


def substitute(thing, callback: Callable[[str], str]):
    if isinstance(thing, (type(None), bool, int, float)):
        return thing
    if isinstance(thing, str):
        fs = re.split(r"<<(\$?[a-zA-Z0-9-_]+)>>", thing)
        result = []
        is_even = True
        for f in fs:
            is_even = not is_even
            if is_even:
                result.append(callback(f))
            else:
                result.append(f)
        return "".join(result)
    elif isinstance(thing, list):
        result = []
        for item in thing:
            result.append(substitute(item, callback))
        return result
    else:
        result = {}
        for key, value in thing.items():
            key = substitute(key, callback)
            value = substitute(value, callback)
            result[key] = value
        return result


def merge(*objects):
    if all(isinstance(item, list) for item in objects):
        result = []
        for ls in objects:
            for item in ls:
                result.append(item)
        return result
    if not all(isinstance(item, dict) for item in objects):
        raise NoMergePossible()
    result = {}
    for obj in objects:
        for key, value in obj.items():
            if key not in result:
                result[key] = value
                continue
            if isinstance(value, list):
                if not isinstance(result[key], list):
                    result[key] = [result[key]]
            if isinstance(result[key], list):
                if isinstance(value, list):
                    for item in value:
                        result[key].append(item)
                    continue
                result[key].append(value)
                continue
            if isinstance(result[key], dict):
                if not isinstance(value, dict):
                    raise NoMergePossible()
                for item_key, item_value in value.items():
                    result[key][item_key] = item_value
                continue
            result[key] = value
    return result


def parse_name(name) -> List[str]:
    parts = []
    for p1 in str(name).split("-"):
        for part in p1.split("_"):
            current_part = []
            last_was_upper = False
            for char in part:
                if char.isupper() and not last_was_upper:
                    if current_part:
                        parts.append("".join(current_part).lower())
                    current_part = []
                elif char.islower() and last_was_upper:
                    last = current_part.pop()
                    if current_part:
                        parts.append("".join(current_part).lower())
                    current_part = [last]
                current_part.append(char)
                last_was_upper = char.isupper()
            if current_part:
                parts.append("".join(current_part).lower())
    return parts


def simplify_dict(thing):
    try:
        items = thing.items()
    except AttributeError:
        return None
    result = {}
    for key, value in items:
        result[key] = simplify(value)
    return result


def simplify(thing):
    if thing is None:
        return thing
    if isinstance(thing, (bool, str, int, float, Decimal)):
        return thing
    if isinstance(thing, Dec):
        if thing.scale == 0:
            return int(thing)
        else:
            return Decimal(str(thing))
    # noinspection PyTypeChecker
    if (result := simplify_dict(thing)) is not None:
        return result
    # noinspection PyTypeChecker
    return [simplify(elem) for elem in thing]


def empty(value) -> bool:
    if isinstance(value, (bool, list, dict, str)):
        return not value
    return value is None



class JinsiException(Exception):
    pass


class NoParseError(JinsiException):
    pass


class MalformedNameError(NoParseError):
    def __init__(self, name: str, expected: str):
        self.name = name
        self.expected = expected

    def __str__(self):
        return f"{self.name}, expected to match: {self.expected}"


class MalformedEachError(NoParseError):
    pass


class NoSuchVariableError(JinsiException):
    def __init__(self, name: str):
        self.name = name


class NoSuchEnvironmentVariableError(JinsiException):
    def __init__(self, name: str):
        self.name = name


class NoSuchFunctionError(NoParseError):
    def __init__(self, name: str):
        self.name = name


class NoMergePossible(JinsiException):
    pass



import re

import yaml
import yaml.composer
import yaml.constructor
import yaml.emitter
import yaml.parser
import yaml.reader
import yaml.representer
import yaml.resolver
import yaml.scanner
import yaml.serializer
from dezimal import Dec


class Resolver(yaml.resolver.BaseResolver):
    pass


Resolver.add_implicit_resolver(
    '!dec',
    re.compile(r'^[-+]?(?:[0-9]+)(?:\.[0-9]+)?(?:[eE][-+][0-9]+)?$'),
    list('-+0123456789')
)

for ch, vs in yaml.resolver.Resolver.yaml_implicit_resolvers.items():
    Resolver.yaml_implicit_resolvers.setdefault(ch, []).extend(
        (tag, regexp) for tag, regexp in vs
        if not tag.endswith('float') and not tag.endswith('int')
    )


class Loader(yaml.reader.Reader, yaml.scanner.Scanner, yaml.parser.Parser,
             yaml.composer.Composer, yaml.constructor.SafeConstructor, Resolver):

    def __init__(self, stream):
        yaml.reader.Reader.__init__(self, stream)
        yaml.scanner.Scanner.__init__(self)
        yaml.parser.Parser.__init__(self)
        yaml.composer.Composer.__init__(self)
        yaml.constructor.SafeConstructor.__init__(self)
        Resolver.__init__(self)


def dec_constructor(loader, node):
    value = loader.construct_scalar(node)
    return Dec(value)


yaml.add_constructor('!dec', dec_constructor, Loader)


def str_node(value: str) -> yaml.ScalarNode:
    return yaml.ScalarNode(tag="tag:yaml.org,2002:str", value=value)


def map_node(value) -> yaml.MappingNode:
    return yaml.MappingNode(tag="tag:yaml.org,2002:map", value=value)


def seq_node(value) -> yaml.SequenceNode:
    return yaml.SequenceNode(tag="tag:yaml.org,2002:seq", value=value)


def aws_cloudformation_intrinsic_function(loader, node):
    fn = f"Fn::{node.tag[1:]}"
    if node.tag in ('!Ref', '!Condition'):
        fn = 'Ref'
    elif node.tag == '!GetAtt' and isinstance(node.value, str):
        path = node.value.split(".", maxsplit=2)
        node = yaml.SequenceNode(
            tag="tag:yaml.org,2002:seq",
            value=[str_node(p) for p in path]
        )
    sub_node: yaml.Node
    if isinstance(node, yaml.SequenceNode):
        sub_node = seq_node(node.value)
    elif isinstance(node, yaml.MappingNode):
        sub_node = map_node(node.value)
    elif node.value is None:
        sub_node = yaml.ScalarNode('tag:yaml.org,2002:null', node.value)
    elif isinstance(node.value, str):
        sub_node = yaml.ScalarNode('tag:yaml.org,2002:str', node.value)
    elif isinstance(node.value, bool):
        sub_node = yaml.ScalarNode('tag:yaml.org,2002:bool', node.value)
    elif isinstance(node.value, int):
        sub_node = yaml.ScalarNode('tag:yaml.org,2002:int', node.value)
    elif isinstance(node.value, float):
        sub_node = yaml.ScalarNode('tag:yaml.org,2002:float', node.value)
    else:
        raise ValueError(node)
    new_node = map_node([(str_node(fn), sub_node)])
    return loader.construct_object(new_node)


aws_cloudformation_intrinsic_functions = [
    "Base64",
    "Cidr",
    "FindInMap",
    "GetAtt",
    "GetAZs",
    "ImportValue",
    "Join",
    "Select",
    "Split",
    "Sub",
    "Transform",
    "Ref",
    "And",
    "Equals",
    "If",
    "Not",
    "Or",
    "Condition",
]

for func in aws_cloudformation_intrinsic_functions:
    yaml.add_constructor(f"!{func}", aws_cloudformation_intrinsic_function, Loader)


class Dumper(yaml.Dumper):

    def represent_str(self, data):
        if '\n' in data:
            return self.represent_scalar('tag:yaml.org,2002:str', data, style='|')
        else:
            return self.represent_scalar('tag:yaml.org,2002:str', data)

    def represent_dec(self, data):
        s = str(data)
        if '.' in s:
            tag = 'tag:yaml.org,2002:float'
        else:
            tag = 'tag:yaml.org,2002:int'
        return self.represent_scalar(tag, str(data))

    def ignore_aliases(self, data):
        return True


yaml.add_representer(Dec, Dumper.represent_dec, Dumper=Dumper)
yaml.add_representer(str, Dumper.represent_str, Dumper=Dumper)


def dumpyaml(data) -> str:
    return yaml.dump(
        data,
        Dumper=Dumper,
        default_flow_style=False,
    )



from __future__ import annotations

import os
from typing import Dict, Any as Value

from .exceptions import NoSuchEnvironmentVariableError


class Environment:
    def __init__(self, **env):
        self.dyn: Dict[str, Value] = {}
        for key, value in env.items():
            self.dyn[key] = value

    @staticmethod
    def get_var(key: str) -> Value:
        return os.getenv(key)

    def get_dyn(self, key: str) -> Value:
        if key in self.dyn:
            return self.dyn[key]
        raise NoSuchEnvironmentVariableError(key)

    def with_env(self, env: Dict[str, Value]) -> Environment:
        new_env = Environment()
        for key, value in self.dyn.items():
            new_env.dyn[key] = value
        for key, value in env.items():
            new_env.dyn[key] = value
        return new_env



from __future__ import annotations

import re
from datetime import date, datetime
from inspect import getattr_static

import yaml
from dezimal import Dec

from .exceptions import MalformedEachError, MalformedNameError, NoParseError, NoSuchFunctionError
from .expressions import parse_expression
from .functions import Functions
from .nodes import *
from .util import merge


# noinspection PyMethodMayBeStatic
class Parser:

    def __init__(self):
        self.name_regex = "^[a-z]([_-]?[a-z0-9])*$"
        self.path = []

    def check_name(self, name):
        if not re.match(self.name_regex, name):
            raise MalformedNameError(name=name, expected=self.name_regex)

    def parse_node(self, obj: Value, parent: Node) -> Node:
        if isinstance(obj, list):
            return self.parse_sequence(obj, parent)
        if isinstance(obj, (type(None), bool, int, float, str, Dec, date, datetime)):
            return self.parse_constant(obj, parent)
        if '::include' in obj:
            includes = obj['::include']
            if isinstance(includes, str):
                includes = [includes]
            if not isinstance(includes, list):
                raise NoParseError()
            del obj['::include']
            docs = [obj]
            for include in includes:
                with open(include) as f:
                    docs.append(yaml.safe_load(f))
            obj = merge(*docs)
        key_set = set(obj.keys())
        if key_set == {'::all'}:
            return self.parse_all(obj['::all'], parent)
        if key_set == {'::any'}:
            return self.parse_any(obj['::any'], parent)
        if key_set == {'::when', '::then'} or key_set == {'::when', '::then', '::else'}:
            return self.parse_conditional(obj, parent)
        for keyword in ['::all', '::any', '::when', '::then']:
            if keyword in key_set:
                raise NoParseError()
        if '::else' in obj:
            return self.parse_else(obj, parent)
        if '::let' in obj:
            return self.parse_let(obj, parent)
        nodes = []
        remaining = {}
        # TODO: decide - if ::when or ::then are here then NoParse - maybe?
        for key, value in obj.items():
            if '::get' in obj:
                nodes.append(parse_expression(obj['::get'], parent))
            elif key[:8] == "::format":
                nodes.append(self.parse_format(value, parent))
            elif key[:6] == "::call":
                nodes.append(self.parse_application(key, value, parent))
            elif key[:6] == "::each":
                nodes.append(self.parse_each(key, value, parent))
            elif len(key) > 2 and key[:2] == "::":
                if key not in {"::get", "::call", "::each", "::format"}:
                    nodes.append(self.parse_function_application(key, value, parent))
            else:
                remaining[key] = value

        if not remaining and len(nodes) == 1:
            return nodes[0]
        if not nodes:
            return self.parse_object(remaining, parent)
        nodes.append(self.parse_object(remaining, parent))
        merge_node = FunctionApplication(parent, function=Functions.merge)
        for node in nodes:
            node.parent = merge_node
        merge_node.args = nodes
        return merge_node

    def parse_constant(self, obj, parent: Node) -> Node:
        return Constant(parent, obj)

    def parse_let(self, obj: Value, parent: Node) -> Node:
        node = Let(parent)
        remainder = {}
        if not isinstance(obj, dict):
            raise NoParseError()
        for key, value in obj.items():
            if key == '::let':
                for var_key, var_value in value.items():
                    if var_key[:1] == "$":
                        var_key = var_key[1:]
                        self.check_name(var_key)
                        node.env[var_key] = self.parse_node(var_value, node)
                    else:
                        self.check_name(var_key)
                        node.let[var_key] = self.parse_node(var_value, node)
                continue
            remainder[key] = value
        node.body = self.parse_node(remainder, node)
        return node

    def parse_conditional(self, obj, parent: Node) -> Node:
        node = When(parent)
        node.when = self.parse_node(obj['::when'], node)
        node.then = self.parse_node(obj['::then'], node)
        if '::else' in obj:
            node.else_ = self.parse_node(obj['::else'], parent)
        return node

    def parse_all(self, obj, parent: Node) -> Node:
        if not isinstance(obj, list):
            raise NoParseError()
        node = All(parent)
        for item in obj:
            node.nodes.append(self.parse_node(item, node))
        return node

    def parse_any(self, obj, parent: Node) -> Node:
        if not isinstance(obj, list):
            raise NoParseError()
        node = Any(parent)
        for item in obj:
            node.nodes.append(self.parse_node(item, node))
        return node

    def parse_else(self, obj: Value, parent: Node) -> Node:
        node = Else(parent)
        remainder = {}
        for key, value in obj.items():
            if key == '::else':
                node.otherwise = self.parse_node(value, node)
                continue
            remainder[key] = value
        node.body = self.parse_node(remainder, node)
        return node

    def parse_object(self, obj, parent: Node) -> Node:
        node = Object(parent)
        for key, value in obj.items():
            node.children[key] = self.parse_node(value, node)
        return node

    def parse_sequence(self, obj, parent: Node) -> Node:
        seq = Sequence(parent)
        for item in obj:
            seq.elements.append(self.parse_node(item, seq))
        return seq

    def parse_function_application(self, key: str, args, parent: Node) -> Node:
        name = key[2:]
        if name[-1:] == "_":
            name = name[:-1]
        try:
            if not isinstance(getattr_static(Functions, name), staticmethod):
                raise NoSuchFunctionError(name)
        except AttributeError:
            raise NoSuchFunctionError(name)
        func = getattr(Functions, name)
        app = FunctionApplication(parent, func)
        if isinstance(args, list):
            for arg in args:
                app.args.append(self.parse_node(arg, app))
        else:
            if isinstance(args, str):
                app.args.append(self.parse_format(args, app))
            else:
                app.args.append(self.parse_node(args, app))
        return app

    def parse_application(self, key: str, value, parent: Node) -> Node:
        if key == "::call" and isinstance(value, str):
            name = value
        else:
            try:
                _, name = key.split(" ")
            except ValueError:
                name = ""
        self.check_name(name)
        app = Application(parent, name)
        if isinstance(value, list):
            app.kwargs[""] = self.parse_sequence(value, app)
        elif isinstance(value, dict):
            for arg_key, arg_value in value.items():
                if arg_key[:1] == "$":
                    arg_key = arg_key[1:]
                app.kwargs[arg_key] = self.parse_node(arg_value, app)
        return app

    def parse_each(self, key: str, value, parent: Node) -> Node:
        try:
            _, source, as_, target = key.split(" ")
            if as_ != "as":
                raise MalformedEachError()
        except ValueError:
            raise MalformedEachError()
        each = Each(parent, source, target)
        each.body = self.parse_node(value, each)
        return each

    def parse_format(self, value, parent: Node) -> Node:
        return Format(parent, value)



import re

from dezimal import Dec

from .functions import Functions
from .nodes import *


def peek(stack):
    return stack[-1] if stack else None


PRECEDENCES = {
    'or': 1,
    'and': 2,
    '==': 3,
    '!=': 3,
    '<': 4,
    '<=': 4,
    '>': 5,
    '>=': 5,
    '+': 6,
    '-': 6,
    '*': 7,
    '/': 7,
}


# noinspection PyDefaultArgument
def greater_precedence(op1, op2, precedences=PRECEDENCES):
    return precedences[op1] > precedences[op2]


FUNCTIONS = {
    'or': Functions.or_,
    'and': Functions.and_,
    '==': Functions.eq,
    '!=': Functions.neq,
    '<': Functions.lt,
    '>': Functions.gt,
    '<=': Functions.lte,
    '>=': Functions.gte,
    '+': Functions.add,
    '-': Functions.sub,
    '*': Functions.mul,
    '/': Functions.div,
}


# noinspection PyDefaultArgument
def build_node(operators: List[str], nodes: List[Node], functions=FUNCTIONS):
    operator = operators.pop()
    right = nodes.pop()
    left = nodes.pop()
    func = functions[operator]
    node = FunctionApplication(parent=Empty(), function=func)
    node.args = [left, right]
    left.parent = node
    right.parent = node
    nodes.append(node)


DIGITS = frozenset(x for x in "0123456789")


def is_digit(char, digits=DIGITS):
    return char in digits


LETTERS = frozenset(x for x in "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")


def is_letter(char, letters=LETTERS):
    return char in letters


def is_env_var(name, regex=re.compile(r"^[0-9_]*[A-Z][_A-Z0-9]*$")):
    return regex.match(name)


def parse_expression(
        expr: str,
        parent: Node,
        regex=re.compile(r"-?[0-9]+|-?[0-9]+\\.[0-9]+|-?\$?[a-zA-Z][a-zA-Z0-9._-]*|[+/*()<>=!-]+"),
) -> Node:
    tokens = regex.findall(expr)
    nodes = []
    operators = []
    for token in tokens:
        if token[0].isdigit() or (len(token) > 1 and token[0] == '-' and token[1].isdigit()):
            nodes.append(Constant(parent=Empty(), value=Dec(token)))
        elif token[0].isalpha() and token not in ('and', 'or'):
            if is_env_var(token):
                nodes.append(GetEnvVar(parent=Empty(), name=token))
            else:
                nodes.append(GetLet(parent=Empty(), path=token.split(".")))
        elif token[0] == '$':
            nodes.append(GetDyn(parent=Empty(), path=token[1:].split(".")))
        elif token == '(':
            operators.append(token)
        elif token == ')':
            top = peek(operators)
            while top is not None and top != '(':
                build_node(operators, nodes)
                top = peek(operators)
            operators.pop()  # Discard the '('
        else:
            top = peek(operators)
            while top is not None and top not in "()" and greater_precedence(top, token):
                build_node(operators, nodes)
                top = peek(operators)
            operators.append(token)
    while peek(operators) is not None:
        build_node(operators, nodes)
    # TODO: handle error if len(nodes) != 1
    result = nodes[0]
    result.parent = parent
    return result



import datetime
import json
from typing import Iterator, Any, Union

from dezimal import Dec

INFINITY = float('inf')


class Encoder(json.JSONEncoder):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    # noinspection PyMethodMayBeStatic
    def default_raw(self, o: Any) -> Union[str, type(None)]:
        return None

    def encode(self, o: Any) -> str:
        pieces = self.iterencode(o)
        return "".join(list(pieces))

    def iterencode(self, obj: Any, **kwargs) -> Iterator[str]:

        def floatstr(
                o,
                allow_nan=self.allow_nan,
                # check for specials. this type of test is platform-specific, so do tests which don't depend internals.
                _repr=repr,
                _inf=INFINITY,
                _neginf=-INFINITY
        ):

            if o != o:
                text = 'NaN'
            elif o == _inf:
                text = 'Infinity'
            elif o == _neginf:
                text = '-Infinity'
            else:
                return _repr(o)
            if not allow_nan:
                raise ValueError(
                    "Out of range float values are not JSON compliant: " +
                    repr(o))
            return text

        return _make_iterencode(
            encode_str=super().encode,
            encode_int=int.__repr__,
            encode_float=floatstr,
            encode_other_raw=self.default_raw,
            encode_other=self.default,
        )(obj)


# noinspection PyShadowingBuiltins
def _make_iterencode(
        encode_str,
        encode_int,
        encode_float,
        encode_other_raw,
        encode_other,
        # HACK: hand-optimized bytecode; turn globals into locals
        iter=iter,
        next=next,
        isinstance=isinstance,
        str=str,
        dict=dict,
        list=list,
        int=int,
        float=float,
        dec=Dec
):
    def _iterencode_dict(o):
        if not o:
            yield '{}'
            return
        yield '{'
        it = iter(o.items())
        key, value = next(it)
        yield encode_str(key)
        yield ':'
        yield from _iterencode(value)
        for key, value in it:
            yield ','
            yield encode_str(key)
            yield ':'
            yield from _iterencode(value)
        yield '}'

    def _iterencode_list(o):
        if not o:
            yield '[]'
            return
        yield '['
        it = iter(o)
        item = next(it)
        yield from _iterencode(item)
        for item in it:
            yield ','
            yield from _iterencode(item)
        yield ']'

    def _iterencode(o):
        if isinstance(o, str):
            yield encode_str(o)
        elif o is None:
            yield 'null'
        elif o is True:
            yield 'true'
        elif o is False:
            yield 'false'
        elif isinstance(o, dict):
            yield from _iterencode_dict(o)
        elif isinstance(o, list):
            yield from _iterencode_list(o)
        elif isinstance(o, dec):
            yield str(o)
        elif isinstance(o, int):
            yield encode_int(o)
        elif isinstance(o, float):
            yield encode_float(o)
        elif isinstance(o, (datetime.date, datetime.datetime)):
            yield encode_str(str(o))
        else:
            raw = encode_other_raw(o)
            if raw is not None:
                yield raw
            else:
                yield from _iterencode(encode_other(o))

    return _iterencode


def dumpjson(obj, **kwargs) -> str:
    return json.dumps(obj, cls=Encoder, **kwargs)



import re
import sys

from jinsi import parse_yaml, parse_file, render_json, render_yaml


def main():
    args = []
    env = {}
    fmt_json = False
    args_it = iter(sys.argv)
    next(args_it)
    for arg in args_it:
        if arg in ("-j", "-json", "--json"):
            fmt_json = True
            continue
        if m := re.match(r"([^=]+)=(.*)", arg):
            key = m.group(1)
            val = m.group(2)
            env[key] = val
            continue
        args.append(arg)
    for arg in args:
        if arg == '-':
            doc = parse_yaml(sys.stdin.read())
        else:
            doc = parse_file(arg)
        if fmt_json:
            print(render_json(doc, **env))
        else:
            print("---")
            print(render_yaml(doc, **env))


main()



from __future__ import annotations

from typing import Dict, List, Any as Value

from .environment import Environment
from .exceptions import NoSuchVariableError, NoSuchEnvironmentVariableError
from .util import Singleton, select, substitute, empty, cached_method


class Node:
    is_empty = False

    def __init__(self, parent: Node):
        self.parent: Node = parent
        self.is_empty = False

    def get_let(self, name: str) -> Node:
        return self.parent.get_let(name)

    def evaluate(self, env: Environment) -> Value:
        return None


class Empty(Node, metaclass=Singleton):
    is_empty = True

    def __init__(self):
        super().__init__(self)

    def get_let(self, name: str) -> Node:
        raise NoSuchVariableError(name)


class Constant(Node):
    def __init__(self, parent: Node, value: Value):
        super().__init__(parent)
        self.value = value

    def evaluate(self, env: Dict[str, Value]) -> Value:
        return self.value


class GetLet(Node):
    def __init__(self, parent: Node, path: List[str]):
        super().__init__(parent)
        self.path: List[str] = path

    def evaluate(self, env: Environment) -> Value:
        result = self.get_let(self.path[0]).evaluate(env)
        if len(self.path) > 1:
            result = select(result, *self.path[1:])
        return result


class GetDyn(Node):
    def __init__(self, parent: Node, path: List[str]):
        super().__init__(parent)
        self.path: List[str] = path

    def evaluate(self, env: Environment) -> Value:
        result = env.get_dyn(self.path[0])
        if len(self.path) > 1:
            result = select(result, *self.path[1:])
        return result


class GetEnvVar(Node):
    def __init__(self, parent: Node, name: str):
        super().__init__(parent)
        self.name: str = name

    def evaluate(self, env: Environment) -> Value:
        return env.get_var(self.name)


class Let(Node):
    def __init__(self, parent: Node):
        super().__init__(parent)
        self.let: Dict[str, Node] = {}
        self.env: Dict[str, Node] = {}
        self.body: Node = Empty()

    def evaluate(self, env: Environment) -> Value:
        if not self.env:
            return self.body.evaluate(env)
        my_env: Dict[str, Value] = {}
        for key, node in self.env.items():
            my_env[key] = node.evaluate(env)
        return self.body.evaluate(env.with_env(my_env))

    def get_let(self, name: str) -> Node:
        if name not in self.let:
            return super().get_let(name)
        return self.let[name]


class Else(Node):
    def __init__(self, parent: Node):
        super().__init__(parent)
        self.body: Node = Empty()
        self.otherwise: Node = Empty()

    def evaluate_body(self, env: Environment):
        try:
            return self.body.evaluate(env)
        except (NoSuchEnvironmentVariableError, ArithmeticError, ValueError, TypeError):
            return None

    def evaluate(self, env: Environment) -> Value:
        result = self.evaluate_body(env)
        if empty(result):
            result = self.otherwise.evaluate(env)
        return result


class Object(Node):
    def __init__(self, parent: Node):
        super().__init__(parent)
        self.children: Dict[str, Node] = {}

    def evaluate(self, env: Environment) -> Value:
        result = {}
        for key, node in self.children.items():
            result[key] = node.evaluate(env)
        return result


class Sequence(Node):
    def __init__(self, parent: Node):
        super().__init__(parent)
        self.elements: List[Node] = []

    def evaluate(self, env: Environment) -> Value:
        result = []
        for element in self.elements:
            result.append(element.evaluate(env))
        return result


class FunctionApplication(Node):
    def __init__(self, parent: Node, function):
        super().__init__(parent)
        self.function = function
        self.args: List[Node] = []

    def evaluate(self, env: Environment) -> Value:
        args = []
        for arg in self.args:
            args.append(arg.evaluate(env))
        result = self.function(*args)
        return result


class Application(Node):
    def __init__(self, parent: Node, template: str):
        super().__init__(parent)
        self.template = template
        self.kwargs: Dict[str, Node] = {}

    @cached_method
    def evaluate(self, env: Environment) -> Value:
        my_env: Dict[str, Value] = {}
        for key, node in self.kwargs.items():
            my_env[key] = node.evaluate(env)
        return self.get_let(self.template).evaluate(env.with_env(my_env))


class Each(Node):
    def __init__(self, parent: Node, source: str, target: str):
        super().__init__(parent)
        self.source: str = source
        self.target: str = target
        self.body: Node = Empty()
        self._value: Node = Empty()

    def get_let(self, name: str) -> Node:
        if name == self.target:
            return self._value
        else:
            return super().get_let(name)

    def evaluate(self, env: Environment) -> Value:
        if self.source[:1] == "$":
            value = env.get_dyn(self.source[1:])
        else:
            value = self.parent.get_let(self.source).evaluate(env)
        results = []
        if self.target[:1] == "$":
            target = self.target[1:]
            for entry in value:
                result = self.body.evaluate(env.with_env({target: entry}))
                results.append(result)
        else:
            for entry in value:
                self._value: Node = Constant(self, entry)
                result = self.body.evaluate(env)
                results.append(result)
        return results


class When(Node):
    def __init__(self, parent: Node):
        super().__init__(parent)
        self.when: Node = Empty()
        self.then: Node = Empty()
        self.else_: Node = Empty()

    def evaluate(self, env: Environment) -> Value:
        if not empty(self.when.evaluate(env)):
            return self.then.evaluate(env)
        else:
            return self.else_.evaluate(env)


class All(Node):
    def __init__(self, parent: Node):
        super().__init__(parent)
        self.nodes: List[Node] = []

    def evaluate(self, env: Environment) -> Value:
        for node in self.nodes:
            if empty(node.evaluate(env)):
                return False
        return True


class Any(Node):
    def __init__(self, parent: Node):
        super().__init__(parent)
        self.nodes: List[Node] = []

    def evaluate(self, env: Environment) -> Value:
        for node in self.nodes:
            result = node.evaluate(env)
            if not empty(result):
                return result
        return False


class Format(Node):
    def __init__(self, parent: Node, value: Value):
        super().__init__(parent)
        self.value: Value = value

    def evaluate(self, env: Environment) -> Value:
        def subst(key: str) -> str:
            if key[:1] == "$":
                return env.get_dyn(key[1:])
            return self.get_let(key).evaluate(env)

        return substitute(self.value, subst)



import hashlib

from dezimal import Dec

from .util import parse_name, empty


class Functions:

    @staticmethod
    def range_inclusive(from_, to):
        return [Dec(x) for x in range(int(from_), int(to) + 1)]

    @staticmethod
    def range_exclusive(from_, to):
        return [Dec(x) for x in range(int(from_), int(to))]

    # text functions

    @staticmethod
    def titlecase(value):
        return "".join(part.capitalize() for part in parse_name(value))

    @staticmethod
    def kebabcase(value):
        return "-".join(parse_name(value))

    @staticmethod
    def snakecase(value):
        return "_".join(parse_name(value))

    @staticmethod
    def camelcase(value):
        name = Functions.titlecase(value)
        return name[:1].lower() + name[1:]

    @staticmethod
    def uppercase(value):
        return str(value).upper()

    @staticmethod
    def lowercase(value):
        return str(value).lower()

    @staticmethod
    def trim(value, chars=None):
        return str(value).strip(chars)

    @staticmethod
    def trim_left(value, chars=None):
        return str(value).lstrip(chars)

    @staticmethod
    def trim_right(value, chars=None):
        return str(value).rstrip(chars)

    @staticmethod
    def pad_left(value, char, total_length):
        if not isinstance(value, str):
            value: str = str(value)
        if not isinstance(char, str):
            char: str = str(char)
        diff: int = int(total_length) - len(value)
        if diff <= 0:
            return value
        return char * diff + value

    @staticmethod
    def pad_right(value, char, total_length):
        if not isinstance(value, str):
            value: str = str(value)
        if not isinstance(char, str):
            char: str = str(char)
        diff: int = int(total_length) - len(value)
        if diff <= 0:
            return value
        return value + char * diff

    @staticmethod
    def explode(separator, value):
        return value.split(separator)

    @staticmethod
    def implode(separator, items):
        if not items:
            return ""
        result = []
        it = iter(items)
        head = next(it)
        result.append(head)
        for item in items:
            result.append(separator)
            result.append(item)
        return "".join(result)

    # hashes

    @staticmethod
    def md5(value):
        return hashlib.new("md5", value).hexdigest()

    @staticmethod
    def sha1(value):
        return hashlib.new("sha1", value).hexdigest()

    @staticmethod
    def sha256(value):
        return hashlib.new("sha256", value).hexdigest()

    @staticmethod
    def sha512(value):
        return hashlib.new("sha512", value).hexdigest()

    @staticmethod
    def sha3_256(value):
        return hashlib.new("sha3_256", value).hexdigest()

    @staticmethod
    def sha3_512(value):
        return hashlib.new("sha3_512", value).hexdigest()

    # value tests

    @staticmethod
    def is_empty(value):
        return empty(value)

    # type tests

    @staticmethod
    def is_null(value):
        return value is None

    @staticmethod
    def is_number(value):
        return isinstance(value, (Dec, int, float))

    @staticmethod
    def is_string(value):
        return isinstance(value, str)

    @staticmethod
    def is_boolean(value):
        return isinstance(value, bool)

    @staticmethod
    def is_list(value):
        return isinstance(value, list)

    @staticmethod
    def is_object(value):
        return isinstance(value, dict)

    # data conversion

    @staticmethod
    def number(value):
        return Dec(value)

    @staticmethod
    def string(value):
        return str(value)

    @staticmethod
    def boolean(value):
        return bool(value)

    @staticmethod
    def list(value):
        if isinstance(value, list):
            return value
        return [value]

    # aggregation

    @staticmethod
    def sum(*args):
        result = Dec(0)
        for arg in args:
            result += Dec(arg)
        return result

    @staticmethod
    def product(*args):
        result = Dec(1)
        for arg in args:
            result *= Dec(arg)
        return result

    # comparison

    @staticmethod
    def eq(a, b):
        return a == b

    @staticmethod
    def neq(a, b):
        return a != b

    @staticmethod
    def lt(a, b):
        return a < b

    @staticmethod
    def gt(a, b):
        return a > b

    @staticmethod
    def lte(a, b):
        return a <= b

    @staticmethod
    def gte(a, b):
        return a >= b

    @staticmethod
    def and_(a, b):
        return a and b

    @staticmethod
    def or_(a, b):
        return a or b

    # arithmetic

    @staticmethod
    def add(a, b):
        return Dec(a) + Dec(b)

    @staticmethod
    def sub(a, b):
        return Dec(a) - Dec(b)

    @staticmethod
    def mul(a, b):
        return Dec(a) * Dec(b)

    @staticmethod
    def div(a, b, maxscale=None, minscale=17):
        return Dec.div(Dec(a), Dec(b), maxscale, minscale)

    # lists and strings

    @staticmethod
    def length(items):
        return len(items)

    @staticmethod
    def select(items, ix):
        return items[ix]

    @staticmethod
    def contains(needle, haystack):
        return needle in haystack

    @staticmethod
    def starts_with(prefix, value):
        for ix in range(0, len(prefix)):
            if prefix[ix] != value[ix]:
                return False
        return True

    @staticmethod
    def ends_with(prefix, value):
        for ix in range(1, len(prefix) + 1):
            if prefix[-ix] != value[-ix]:
                return False
        return True

    @staticmethod
    def reverse(items):
        return reversed(items)

    @staticmethod
    def sort(items):
        return sorted(items)

    @staticmethod
    def take(n, items):
        if not isinstance(items, list):
            items = str(items)
        return items[:int(n)]

    @staticmethod
    def drop(n, items):
        if not isinstance(items, list):
            items = str(items)
        return items[int(n):]

    @staticmethod
    def concat(*items):
        if all(isinstance(item, list) for item in items):
            result = []
            for ls in items:
                for item in ls:
                    result.append(item)
        return "".join([str(s) for s in items])

    # object creation utilities

    @staticmethod
    def object(key, value):
        return {key: value}

    @staticmethod
    def merge(*items):
        obj = {}
        for item in items:
            if item is None:
                continue
            if isinstance(item, list):
                item = Functions.merge(*item)
            for key, value in item.items():
                obj[key] = value
        return obj
